//
// Open Service Platform
// Copyright (c) 2012 Samsung Electronics Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the License);
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

/**
 * @file	FIo_ChannelWebAppStub.h
 * @brief	This is the header file for the _ChannelWebAppStub class.
 *
 * This file contains the declarations of _ChannelWebAppStub.
 */

#ifndef _FIO_INTERNAL_CHANNEL_WEBAPP_STUB_H_
#define _FIO_INTERNAL_CHANNEL_WEBAPP_STUB_H_

#include <string>
#include <pthread.h>
#include <map>
#include <glib.h>
#include <libsoup/soup.h>

#include <FOspConfig.h>
#include <FBaseObject.h>
#include <FBaseString.h>
#include <FBaseResult.h>
#include <FBaseColArrayList.h>
#include <FBaseColHashMap.h>
#include <FAppTypes.h>
#include <FIo_ChannelService.h>

namespace Tizen { namespace Base { namespace Runtime
{
class _EventDispatcher;
}}}

namespace Tizen { namespace Io
{

enum WebSocketType
{
	WS_NONE,
	WS_HYBI00,
	WS_HYBI10
};

class _IIpcServerEventListener;
class _ChannelServiceStub;

/**
 * @class _ChannelWebAppStub
 * @brief This class provides methods to handle web-application request messages.
 * @since 2.1
 *
 */
class _OSP_EXPORT_ _ChannelWebAppStub
	: public Tizen::Base::Object
{
public:
	/**
	 * This is the default constructor for this class.
	 * @since 2.1
	 */
	_ChannelWebAppStub(void);

	/**
	 * This is the destructor for this class.
	 * @since 2.1
	 */
	virtual ~_ChannelWebAppStub(void);

	result Construct(void);

	void SetChannelService(_ChannelService& service);

	static bool SendResponse(int clientId, const Tizen::Base::Collection::ArrayList& args);

private:
	_ChannelWebAppStub(const _ChannelWebAppStub& rhs);

	_ChannelWebAppStub& operator =(const _ChannelWebAppStub& rhs);

	static gboolean OnConnectionRequest(GIOChannel* source, GIOCondition condition, gpointer data);

	static gboolean OnReadMessage(GIOChannel* source, GIOCondition condition, gpointer data);

	gboolean HandleReceivedMessage(GIOChannel* source, GIOCondition condition, gpointer data);

	static std::string CreateResponseHeader(int appSocket, SoupMessageHeaders *header, const std::string& response);

	static WebSocketType VerifyHttpHeaderType(std::string header, int headerLength);
	static std::string MakeHybi00Response(uint32_t first, uint32_t second, unsigned char* third);
	static std::string MakeHybi10Response(std::string& secureAccept);

	static bool AppendChallengeNumber(uint32_t number, unsigned char *response);
	static unsigned int ExtractChallengeNumber(const char* keyString);
	static char* EncodeBase64(unsigned char *string, int len);


//	struct  _ClientInfo;
//
//	struct  _ChannelInfo
//	{
//		_ChannelInfo(void);
//		~_ChannelInfo(void);
//
//		struct _ClientInfo* pClientInfo;
//		GIOChannel* pGIOChannel;
//		GSource* pGSource;
//		bool destroySource;
//	};
//
//	struct  _ClientInfo
//	{
//		_ClientInfo(void);
//		~_ClientInfo(void);
//
//		int clientId;                              // the client id
//		_ChannelWebAppStub* pChannelStub;                   // the pointer to an _ChannelWebAppStub
//		//GIOChannel* pReverseChannel;               // the channel for sending reverse message
//		std::vector <struct _ChannelInfo*> channels;   // the set of channels associated with a client
//		//Tizen::Base::String appId;
//		//Tizen::Base::String appExecutableName;
//	};

	class _ClientInfo;

	class _ChannelInfo
	{
	public:
		_ChannelInfo(void);
		~_ChannelInfo(void);

		_ClientInfo* pClientInfo;

		Tizen::App::AppId destAppId;

		GIOChannel* pGIOChannel;
		GSource* pGSource;
		bool destroySource;
	};

	class _ClientInfo
		: public Tizen::Base::Object
	{
	public:
		_ClientInfo(void);
		~_ClientInfo(void);

		Tizen::App::AppId srcAppId;

		//int clientId;
		Tizen::Base::String key; // key = srcAppId + srcAppExecName
		_ChannelWebAppStub* pChannelStub;
		std::vector <struct _ChannelInfo*> channels;   /**< the set of channels associated with a client */
	};

	class _WebSocketMessage
	{
		public:
			_WebSocketMessage(void);
			~_WebSocketMessage(void);


			typedef union
			{
				struct
				{
					unsigned int opcode : 4;
					unsigned int rsv3 : 1;
					unsigned int rsv2 : 1;
					unsigned int rsv1 : 1;
					unsigned int fin : 1;
					unsigned int payloadLen : 7;
					unsigned int masked : 1;
				} factors;
				uint16_t header;
			} MiniHeader;

			int ParseHeader(uint8_t* header, int messageLen);
			int ParsePayload(uint8_t* payload, int messageLen);
			void ConsoleMessage() const;
			static size_t CreateMessage(std::string& payload, std::string& message);

			MiniHeader m_header;
			uint8_t m_payloadOffset;
			uint8_t m_maskingKey[4];
			uint64_t m_payloadLen;
			bool m_headerParsed;
			uint8_t* m_payload;

		private:
			_WebSocketMessage(_WebSocketMessage& webSocketMessage);

			_WebSocketMessage& operator =(const _WebSocketMessage& webSocketMessage);

	};

	_ChannelService* __pChannelService;

	GMainContext* __pGMainContext;
	GSource* __pServerGSource;
	GSource* __pClientGSource;

	Tizen::Base::Collection::HashMap* __pClients;

}; // _ChannelWebAppStub

}} // Tizen::Io

#endif // _FIO_INTERNAL_CHANNEL_WEBAPP_STUB_H_

